from .netbox_models import (
    CreatePrefixInput,
    NetboxCommonArgs,
    NetboxFastApiArgs,
    CreateDeviceInterfacesInput,
)

__all__ = (
    CreatePrefixInput,
    NetboxCommonArgs,
    NetboxFastApiArgs,
    CreateDeviceInterfacesInput,
)
